import pyautogui as gui

from pyautoscene import ImageElement, Scene, Session, TextElement
from pyautoscene.utils import locate_and_click

login = Scene(
    "Login",
    elements=[
        TextElement("Welcome to Login"),
        ImageElement("examples/saucedemo/references/login_button.png"),
    ],
    initial=True,
)

dashboard = Scene(
    "Dashboard",
    elements=[
        TextElement("Swag Labs"),
        TextElement("Products"),
        ImageElement("examples/saucedemo/references/cart_icon.png"),
    ],
)

cart = Scene(
    "Cart",
    elements=[
        TextElement("Your Cart"),
        ImageElement("examples/saucedemo/references/cart_icon.png"),
    ],
)


@login.action(transitions_to=dashboard)
def perform_login(username: str, password: str):
    """Performs the login action to transition from Login to Dashboard."""
    locate_and_click("examples/saucedemo/references/username.png")
    gui.write(username, interval=0.1)
    gui.press("tab")
    gui.write(password, interval=0.1)
    gui.press("enter")


@dashboard.action()
def add_products_to_cart(target: str):
    """Adds products to the cart."""
    locate_and_click(f"examples/saucedemo/references/{target}.png")
    locate_and_click("examples/saucedemo/references/add_to_cart_button.png")


@dashboard.action(transitions_to=cart)
def view_cart():
    """Views the cart."""
    locate_and_click("examples/saucedemo/references/cart_icon.png")


@cart.action()
def checkout():
    """Checks out the items in the cart."""
    locate_and_click("examples/saucedemo/references/checkout_button.png")


session = Session(scenes=[login, dashboard, cart])

# test internal state machine
gui.hotkey("alt", "tab")
print(f"{session.current_scene=}")
session.invoke("perform_login", username="standard_user", password="secret_sauce")
print(f"After login: {session._sm.current_state=}")
session.invoke("add_products_to_cart", target="backpack")
session.invoke("view_cart")
print(f"After adding to cart: {session._sm.current_state=}")
session.invoke("checkout")



import logging
import os
from hashlib import sha256
from pathlib import Path

import numpy as np
from PIL import Image

from .region import Region

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

try:
    from rapidocr import RapidOCR
    from rapidocr.utils.output import RapidOCROutput
except ImportError:
    raise ImportError(
        "RapidOCR is not installed. Please install it using 'pip install pyautoscene[ocr]'."
    )

default_ocr_config_path = Path(__file__).parent / "ocr_config.yaml"
ocr_config_path = Path(os.getenv("PYAUTOSCENE_OCR_CONFIG", default_ocr_config_path))
logger.info(f"OCR config path: {ocr_config_path}")


def hash_image(img: Image.Image) -> str:
    return sha256(img.tobytes()).hexdigest()


def convert_points_to_ltwh(points: np.ndarray) -> Region:
    if points.shape[0] == 0:
        raise ValueError("Points array is empty")

    x_min = np.min(points[:, 0])
    y_min = np.min(points[:, 1])
    x_max = np.max(points[:, 0])
    y_max = np.max(points[:, 1])

    return Region(left=x_min, top=y_min, width=x_max - x_min, height=y_max - y_min)


class OCR:
    engine: RapidOCR | None = None
    img_cache: dict[str, tuple[tuple[str, Region], ...]] = {}

    def __new__(cls):
        if cls.engine is None:
            cls.engine = RapidOCR(config_path=ocr_config_path.as_posix())
        return super().__new__(cls)

    def recognize_text(self, img: Image.Image) -> tuple[tuple[str, Region], ...]:
        img_gray = img.convert("L")
        img_hash = hash_image(img_gray)
        if img_hash in self.img_cache:
            logger.debug(f"Using cached result for image hash: {img_hash}")
            return self.img_cache[img_hash]

        assert self.engine is not None, "Engine should be initialized in __new__"
        result = self.engine(np.array(img_gray))
        assert isinstance(result, RapidOCROutput), (
            "Result should be of type RapidOCROutput"
        )
        assert result.txts is not None and result.boxes is not None, (
            "Text recognition failed, txts and boxes should not be None"
        )

        detections = tuple(
            (txt, convert_points_to_ltwh(box))
            for txt, box in zip(result.txts, result.boxes)
        )
        self.img_cache[img_hash] = detections
        return detections



import os

LOCATE_AND_CLICK_DELAY = float(os.getenv("PYAUTOSCENE_LOCATE_AND_CLICK_DELAY", 0.3))

# pixels per second, used for calculating move duration
POINTER_SPEED = int(os.getenv("PYAUTOSCENE_POINTER_SPEED", 1000))



import pyautogui as gui

from pyautoscene import ImageElement, Scene, Session, TextElement

# Define all elements separately
add_to_cart = ImageElement("examples/saucedemo/references/add_to_cart_button.png")
username = ImageElement("examples/saucedemo/references/username.png")
backpack = TextElement(
    "Sauce Labs Bike Light", region="x:2/2 y:(2-4)/5", case_sensitive=False
)
cart_icon = ImageElement("examples/saucedemo/references/cart_icon.png")
checkout_button = ImageElement("examples/saucedemo/references/checkout_button.png")

login = Scene(
    "Login",
    elements=[
        TextElement("Username", region="x:2/3 y:(1-2)/3"),
        TextElement("Password", region="x:2/3 y:(1-2)/3"),
        # ReferenceImage("examples/saucedemo/references/login_button.png"),
    ],
    initial=True,
)

dashboard = Scene(
    "Dashboard",
    elements=[
        TextElement("Swag Labs", region="x:2/3 y:1/3"),
        TextElement("Products", region="x:1/3 y:1/3"),
    ],
)

cart = Scene(
    "Cart", elements=[TextElement("Your Cart", region="x:1/3 y:1/3"), cart_icon]
)


@login.action(transitions_to=dashboard)
def perform_login(username_inp: str, password_inp: str):
    """Performs the login action to transition from Login to Dashboard."""
    username.locate_and_click()
    gui.write(username_inp, interval=0.1)
    gui.press("tab")
    gui.write(password_inp, interval=0.1)
    gui.press("enter")


@dashboard.action()
def add_products_to_cart(target: str):
    """Adds products to the cart."""
    if target == "backpack":
        backpack.locate_and_click()
    else:
        backpack.locate_and_click()
    add_to_cart.locate_and_click(region="x:2/3 y:(2-3)/3", clicks=1)


@dashboard.action(transitions_to=cart)
def view_cart():
    """Views the cart."""
    cart_icon.locate_and_click()


@cart.action()
def checkout():
    """Checks out the items in the cart."""
    checkout_button.locate_and_click()


session = Session(scenes=[login, dashboard, cart])

gui.hotkey("alt", "tab")
session.expect(dashboard, username_inp="standard_user", password_inp="secret_sauce")
session.invoke("add_products_to_cart", target="backpack")
session.invoke("view_cart")
session.invoke("checkout")



from __future__ import annotations

import logging
import os
import time
from keyword import iskeyword
from typing import Callable, overload

import numpy as np
import pyautogui as gui
from PIL import Image

from ._types import MouseButton, TowardsDirection
from .constants import LOCATE_AND_CLICK_DELAY, POINTER_SPEED
from .region import Region, RegionSpec

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def locate_and_click(
    reference: Image.Image | str,
    clicks: int = 1,
    button: MouseButton = "left",
    region: RegionSpec | None = None,
    confidence: float = 0.999,
    grayscale: bool = True,
    limit: int = 1,
    offset: tuple[int, int] = (0, 0),
    towards: TowardsDirection | None = None,
):
    time.sleep(LOCATE_AND_CLICK_DELAY)
    found_region = locate_on_screen(
        reference,
        region=region,
        confidence=confidence,
        grayscale=grayscale,
        limit=limit,
    )
    assert found_region is not None, f"Could not locate {reference} on screen."
    move_and_click(
        found_region, clicks=clicks, button=button, offset=offset, towards=towards
    )
    time.sleep(LOCATE_AND_CLICK_DELAY)


def move_and_click(
    target_region: RegionSpec,
    clicks: int = 1,
    button: MouseButton = "left",
    offset: tuple[int, int] = (0, 0),
    towards: TowardsDirection | None = None,
):
    """Move to the center or edge of the region and click.

    The offset is always added to the calculated target point.
    For example, for 'bottom', offset=(0, 5) means 5 pixels below the bottom edge.
    """
    _target_region = Region.from_spec(target_region)
    base_points = {
        "top": (_target_region.center[0], _target_region.top),
        "left": (_target_region.left, _target_region.center[1]),
        "bottom": (
            _target_region.center[0],
            _target_region.top + _target_region.height - 1,
        ),
        "right": (
            _target_region.left + _target_region.width - 1,
            _target_region.center[1],
        ),
        None: _target_region.center,
    }
    if towards not in base_points:
        raise ValueError(f"Invalid direction: {towards}")
    base = base_points[towards]
    target = (base[0] + offset[0], base[1] + offset[1])

    current = gui.position()
    duration = np.linalg.norm(np.array(target) - np.array(current)) / POINTER_SPEED
    gui.moveTo(*target, float(duration), gui.easeInOutQuad)  # type: ignore
    gui.click(clicks=clicks, button=button)


def is_valid_variable_name(name):
    return name.isidentifier() and not iskeyword(name)


@overload
def locate_on_screen(
    reference: Image.Image | str,
    region: RegionSpec | None = None,
    confidence: float = 0.999,
    grayscale: bool = True,
    limit: int = 1,
    locator: Callable[[Image.Image, Image.Image], list[Region]] | None = None,
) -> Region | None: ...
@overload
def locate_on_screen(
    reference: Image.Image | str,
    region: RegionSpec | None = None,
    confidence: float = 0.999,
    grayscale: bool = True,
    limit: int = 1,
    locator: Callable[[Image.Image, Image.Image], list[Region]] | None = None,
) -> list[Region] | None: ...
def locate_on_screen(
    reference: Image.Image | str,
    region: RegionSpec | None = None,
    confidence: float = 0.999,
    grayscale: bool = True,
    limit: int = 1,
    locator: Callable[[Image.Image, Image.Image], list[Region]] | None = None,
):
    """Locate a region on the screen."""
    if isinstance(reference, str):
        if not os.path.exists(reference):
            raise FileNotFoundError(f"Image file {reference} does not exist.")
        reference = Image.open(reference)
    if locator is None:
        try:
            location = gui.locateOnScreen(
                reference,
                region=Region.from_spec(region).to_box() if region else None,
                grayscale=grayscale,
                confidence=confidence,
                limit=limit,
            )
            if location:
                return Region.from_box(location)
        except gui.ImageNotFoundException:
            return None
        except FileNotFoundError:
            return None
    else:
        screenshot = gui.screenshot(
            region=Region.from_spec(region).to_box() if region else None
        )
        logger.info(
            f"Searching in region: {Region.from_spec(region).to_box() if region else None}.\nGiven region: {region}"
        )
        detections = locator(reference, screenshot)
        logger.info("total detections: %d", len(detections))
        if len(detections) == 0:
            return None
        elif limit > 1:
            return detections[:limit]
        else:
            return detections[0]



from abc import ABC, abstractmethod
from typing import Callable, override

import pyautogui as gui
from PIL import Image

from ._types import MouseButton, TowardsDirection
from .region import Region, RegionSpec
from .utils import locate_on_screen, move_and_click


class ReferenceElement(ABC):
    """Base class for reference elements used to identify scenes."""

    @abstractmethod
    def locate(self, region: RegionSpec | None = None) -> Region | None:
        """Detect the presence of the reference element."""
        raise NotImplementedError("Subclasses must implement this method")

    def locate_and_click(
        self,
        offset: tuple[int, int] = (0, 0),
        region: RegionSpec | None = None,
        clicks: int = 1,
        button: MouseButton = "left",
        towards: TowardsDirection = None,
    ):
        """Locate the reference element and click on it."""
        region = self.locate(region=region)
        assert region is not None, f"Element {self} not found on screen"
        move_and_click(
            target_region=region,
            clicks=clicks,
            button=button,
            offset=offset,
            towards=towards,
        )


class ImageElement(ReferenceElement):
    """Reference element that identifies a scene by an image."""

    def __init__(
        self,
        path: str | list[str],
        confidence: float = 0.999,
        region: RegionSpec | None = None,
        locator: Callable[[Image.Image, Image.Image], list[Region]] | None = None,
    ):
        self.path = path
        self.confidence = confidence
        self.region = region
        self.locator = locator

    @override
    def locate(self, region: RegionSpec | None = None) -> Region | None:
        """Method to detect the presence of the image in the current screen."""
        if isinstance(self.path, str):
            path = [self.path]  # Ensure path is a list for consistency
        else:
            path = self.path
        for image_path in path:
            try:
                location = locate_on_screen(
                    image_path,
                    region=region if region else self.region,
                    confidence=self.confidence,
                    locator=self.locator,
                )
                return location
            except gui.ImageNotFoundException:
                continue


class TextElement(ReferenceElement):
    """Reference element that identifies a scene by text."""

    def __init__(
        self, text: str, region: RegionSpec | None = None, case_sensitive: bool = False
    ):
        self.text = text
        self.region = region
        self.case_sensitive = case_sensitive
        if not case_sensitive:
            self.text = self.text.lower()

    def locate(self, region: RegionSpec | None = None):
        """Method to detect the presence of the text in the current screen."""
        from .ocr import OCR

        ocr = OCR()
        region = region or self.region
        for text, detected_region in ocr.recognize_text(
            gui.screenshot(region=Region.from_spec(region).to_box() if region else None)
        ):
            if not self.case_sensitive:
                text = text.lower()
            if text.strip() == self.text.strip():
                return detected_region.resolve(base=region)
        return None



from __future__ import annotations

from typing import Callable, TypedDict

from statemachine import State

from pyautoscene.utils import is_valid_variable_name

from .references import ReferenceElement
from .region import Region


class ActionInfo(TypedDict):
    """Type definition for action information in a scene."""

    action: Callable[..., None]
    transitions_to: Scene | None


class Scene(State):
    """A scene represents a state in the GUI automation state machine."""

    def __init__(
        self,
        name: str,
        elements: list[ReferenceElement] | None = None,
        initial: bool = False,
    ):
        assert is_valid_variable_name(name), (
            f"Invalid scene name: {name}, must be a valid Python identifier."
        )
        super().__init__(name, initial=initial)
        self.elements = elements or []
        self.actions: dict[str, ActionInfo] = {}

    def action(self, transitions_to: Scene | None = None):
        """Decorator to register an action for this scene."""

        def decorator(func: Callable[..., None]) -> Callable[..., None]:
            if func.__name__ not in self.actions:
                action_name = func.__name__
                self.actions[action_name] = {
                    "action": func,
                    "transitions_to": transitions_to,
                }
            return func

        return decorator

    def get_action(self, action_name: str) -> ActionInfo | None:
        """Get an action by name."""
        return self.actions.get(action_name)

    def is_on_screen(self, region: Region | None = None) -> bool:
        """Check if any reference element is currently on screen."""
        # TODO: Refactor after text recognition is implemented
        # elements = (elem for elem in self.elements if isinstance(elem, ReferenceImage))
        return all(elem.locate(region) for elem in self.elements)

    def __repr__(self):
        return f"Scene({self.name!r}, elements={len(self.elements)})"



from __future__ import annotations

import time
from typing import Callable

import networkx as nx
from PIL import Image
from statemachine import State, StateMachine
from statemachine.factory import StateMachineMetaclass
from statemachine.states import States
from statemachine.transition_list import TransitionList

from .references import ImageElement, ReferenceElement
from .region import Region
from .scene import Scene


class SceneRecognitionError(Exception):
    pass


def build_dynamic_state_machine(
    scenes: list[Scene],
) -> tuple[StateMachine, dict[str, TransitionList], dict[str, Callable]]:
    """Create a dynamic StateMachine class from scenes using StateMachineMetaclass."""

    states = {scene.name: scene for scene in scenes}
    transitions = {}
    leaf_actions = {}
    for scene in scenes:
        for action_name, action_info in scene.actions.items():
            target_scene = action_info["transitions_to"]
            if target_scene is not None:
                event_name = f"event_{action_name}"
                new_transition = scene.to(target_scene, event=event_name)
                new_transition.on(action_info["action"])
                transitions[event_name] = new_transition
            else:
                leaf_actions[action_name] = action_info["action"]

    SessionSM = StateMachineMetaclass(
        "SessionSM",
        (StateMachine,),
        {"states": States(states), **transitions},  # type: ignore[call-arg]
    )
    session_sm: StateMachine = SessionSM()  # type: ignore[no-redef]

    return session_sm, transitions, leaf_actions


def get_current_scene(scenes: list[Scene], region: Region | None = None) -> Scene:
    """Get the current scene from the list of scenes."""
    current_scenes = [scene for scene in scenes if scene.is_on_screen(region)]
    if len(current_scenes) == 1:
        return current_scenes[0]
    elif len(current_scenes) > 1:
        raise SceneRecognitionError(
            f"Multiple scenes are currently on screen.\n{' '.join(str(scene) for scene in current_scenes)}"
        )
    else:
        raise SceneRecognitionError("No scene is currently on screen.")


class Session:
    """A session manages the state machine for GUI automation scenes."""

    def __init__(
        self,
        scenes: list[Scene],
        image_locator: Callable[[Image.Image, Image.Image], list[Region]] | None = None,
    ):
        self._scenes_list = scenes
        self._scenes_dict = {scene.name: scene for scene in scenes}
        self.image_locator = image_locator
        for scene in self._scenes_list:
            for elem in scene.elements:
                if isinstance(elem, ImageElement):
                    elem.locator = image_locator

        # Create dynamic StateMachine class and instantiate it
        self._sm, self.transitions, self.leaf_actions = build_dynamic_state_machine(
            scenes
        )
        self.graph: nx.MultiDiGraph = nx.nx_pydot.from_pydot(self._sm._graph())

    @property
    def current_scene(self) -> State:
        """Get the current state."""
        return self._sm.current_state

    def expect(self, target_scene: Scene, **kwargs):
        """Navigate to a specific scene."""
        if target_scene.is_on_screen():
            return

        present_scene = get_current_scene(self._scenes_list)
        all_paths = list(
            nx.all_simple_paths(
                self.graph, source=present_scene.name, target=target_scene.name
            )
        )
        if len(all_paths) == 0:
            raise SceneRecognitionError(
                f"No path found from {present_scene.name} to {target_scene.name}"
            )
        elif len(all_paths) > 1:
            raise SceneRecognitionError(
                f"Multiple paths found from {present_scene.name} to {target_scene.name}"
            )

        path = all_paths[0]
        events: list[str] = [
            self.graph.get_edge_data(path[i], path[i + 1])[0]["label"]  # type: ignore
            for i in range(len(path) - 1)
        ]

        for event in events:
            self._sm.send(event, **kwargs)

    def invoke(self, action_name: str, **kwargs):
        """Invoke an action in the current scene."""
        event_name = f"event_{action_name}"
        transition = next(
            (tr for tr_name, tr in self.transitions.items() if tr_name == event_name),
            None,
        )
        if transition:
            return self._sm.send(event_name, **kwargs)

        leaf_action = next(
            (
                action
                for name, action in self.leaf_actions.items()
                if name == action_name
            ),
            None,
        )
        if leaf_action:
            return leaf_action(**kwargs)

        raise ValueError(
            f"Action '{action_name}' not found in current scene '{self.current_scene.name}'"
        )

    def wait_until(self, target: Scene | ReferenceElement, interval: float = 1):
        """Wait until the target scene or reference element is on screen."""
        found = False
        while not found:
            if isinstance(target, Scene):
                found = target.is_on_screen()
            elif isinstance(target, ReferenceElement):
                found = target.locate() is not None
            else:
                raise TypeError("Target must be a Scene or ReferenceElement.")
            if not found:
                time.sleep(interval)

    def __repr__(self):
        current = self.current_scene
        current_name = current.name if current else "None"
        return (
            f"Session(scenes={list(self._scenes_dict.keys())}, current={current_name})"
        )



from typing import Literal

import cv2
import numpy as np
from PIL import Image

from pyautoscene.region import Region

type OpenCvTransformKernel = Literal["CROSS", "RECT", "ELLIPSE"]
type OpenCvTransformMorphology = Literal[
    "ERODE", "DILATE", "OPEN", "CLOSE", "GRADIENT", "TOPHAT", "BLACKHAT", "HITMISS"
]
type OpenCvThresholdAlgo = Literal[
    "BINARY",
    "BINARY_INV",
    "TRUNC",
    "TOZERO",
    "TOZERO_INV",
    "MASK",
    "OTSU",
    "TRIANGLE",
    "DRYRUN",
]
type OpenCvMatchAlgo = Literal[
    "SQDIFF", "SQDIFF_NORMED", "CCORR", "CCORR_NORMED", "CCOEFF", "CCOEFF_NORMED"
]


def ilish(
    needle: Image.Image,
    haystack: Image.Image,
    *,
    transform_kernel: OpenCvTransformKernel = "ELLIPSE",
    transform_shape: tuple[int, int] = (53, 53),
    transform_morphology: OpenCvTransformMorphology = "BLACKHAT",
    threshold: int = 127,
    threshold_max_value: int = 255,
    threshold_algo: OpenCvThresholdAlgo = "BINARY_INV",
    match_algo: OpenCvMatchAlgo = "SQDIFF",
    match_mask_value: int | None = 255,
    confidence: float = 0.9,
):
    def transform(inp: Image.Image) -> np.ndarray:
        gray = cv2.cvtColor(np.array(inp.convert("RGB")), cv2.COLOR_RGB2GRAY)
        kernel = cv2.getStructuringElement(
            getattr(cv2, f"MORPH_{transform_kernel}"), transform_shape
        )
        blackhat = cv2.morphologyEx(
            gray, getattr(cv2, f"MORPH_{transform_morphology}"), kernel
        )
        thresh = cv2.threshold(
            blackhat,
            threshold,
            threshold_max_value,
            getattr(cv2, f"THRESH_{threshold_algo}"),
        )[1]
        return thresh

    if haystack.size[0] < needle.size[0] or haystack.size[1] < needle.size[1]:
        raise ValueError(
            "needle dimension(s) exceed the haystack image or region dimensions"
        )

    needle_t = transform(needle)
    haystack_t = transform(haystack)
    result = cv2.matchTemplate(
        haystack_t,
        needle_t,
        getattr(cv2, f"TM_{match_algo}"),
        None,
        (needle_t != match_mask_value).astype(np.uint8)
        if match_mask_value is not None
        else None,
    )

    if match_algo.startswith("SQDIFF"):
        match_filter = result < confidence
    else:
        match_filter = result > confidence

    match_indices = np.arange(result.size)[match_filter.flatten()]
    matches = np.unravel_index(match_indices, result.shape)

    match_regions = [
        Region(x, y, needle.width, needle.height)
        for x, y in zip(matches[1], matches[0])
    ]
    return (match_regions, result, needle_t, haystack_t)



from __future__ import annotations

import re
from dataclasses import dataclass

import numpy as np
import pyautogui as gui
from pyscreeze import Box

type RegionSpec = Region | str

axis_pattern = re.compile(r"(?P<d>[xy]):\(?(?P<i>\d+)(?:-(?P<j>\d+))?\)?/(?P<n>\d+)")


@dataclass(frozen=True, slots=True)
class Region:
    left: int
    top: int
    width: int
    height: int

    def to_box(self) -> Box:
        """Convert to a pyscreeze Box."""
        return Box(self.left, self.top, self.width, self.height)

    @classmethod
    def from_box(cls, box: Box) -> Region:
        """Create a Region from a pyscreeze Box."""
        return cls(left=box.left, top=box.top, width=box.width, height=box.height)

    @property
    def center(self) -> tuple[int, int]:
        """Get the center coordinates of the region."""
        return (self.left + self.width // 2, self.top + self.height // 2)

    @classmethod
    def from_spec(
        cls, spec: RegionSpec, shape: tuple[int, int] | None = None
    ) -> Region:
        if isinstance(spec, Region):
            return spec
        if shape is None:
            img = np.array(gui.screenshot())
            shape = (img.shape[0]), (img.shape[1])

        default_region = {"left": 0, "top": 0, "width": shape[1], "height": shape[0]}

        axis_mapping = {"x": ("left", "width", 1), "y": ("top", "height", 0)}
        for axis, i, j, n in axis_pattern.findall(spec):
            alignment, size_attr, dim_index = axis_mapping[axis]
            size = shape[dim_index] // int(n)
            i, j = int(i), int(j) if j else int(i)
            default_region.update({
                alignment: (i - 1) * size,
                size_attr: (j - i + 1) * size,
            })

        return cls(**default_region)

    def resolve(self, base: RegionSpec | None) -> Region:
        if base is None:
            return self
        if isinstance(base, str):
            base = Region.from_spec(base)
        return Region(
            left=self.left + base.left,
            top=self.top + base.top,
            width=self.width,
            height=self.height,
        )



from .references import ImageElement, TextElement
from .region import Region, RegionSpec
from .scene import Scene
from .session import Session

__all__ = ["Scene", "Session", "ImageElement", "TextElement", "RegionSpec", "Region"]
